/*
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://eu-egee.org/partners/ for details on the copyright holders.
 *  For license conditions see the license file or http://eu-egee.org/license.html
 *
 *  GLite SrmUtil Class
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>
 *           Peter Kunszt <peter.kunszt@cern.ch>
 *  Version info: $Id: srmUtil.hxx,v 1.12 2006/01/12 10:25:47 badino Exp $
 *  Release: $Name: glite-data-srm-cli_R_1_2_5 $
 * 
 *  Comment: This class originates from io-resolve-common/src/SrmClient
 *           [peter.kunszt@cern.ch]
 */

#ifndef GLITE_DATA_IO_SRM_CLIENT_SRMUTIL_HXX_
#define GLITE_DATA_IO_SRM_CLIENT_SRMUTIL_HXX_

#include <string>
#include "glite/data/srm/cpp/srmISRMProxy.h"

/* Log4cxx pollutes namespace... */
#undef PACKAGE
#undef PACKAGE_NAME
#undef PACKAGE_STRING
#undef PACKAGE_TARNAME
#undef PACKAGE_VERSION
#undef VERSION

#include <log4cxx/logger.h>

#define SRM_LOGGER "glite.data.srm.util"

/*
 * GSOAP Name Wrappers. This code is needed if you want to compile with 
 * GSOAP 2.3 and 2.6
 */
//#if GSOAP_WSDL2H_VERSION < 0x020600
//#   ifndef GSOAP_RETVAL
//#       define GSOAP_RETVAL(a) &a
//#   endif
//#   ifndef GSOAP_FIELDVAL
//#       define GSOAP_FIELDVAL(a,b) a.b
//#   endif
//#else
#   ifndef GSOAP_RETVAL
#       define GSOAP_RETVAL(a) a
#   endif
#   ifndef GSOAP_FIELDVAL
#       define GSOAP_FIELDVAL(a,b) a->b
#   endif
//#endif

namespace srm  {

typedef struct srm2__RequestStatus* SrmStatus;

/**
 * SrmUtil: Common SRM client functionality
 */
class SrmUtil {

public:
    
    /**
     * File Status Pending
     */
    static const char * FILE_STATUS_PENDING;

    /**
     * File Status Ready
     */
    static const char * FILE_STATUS_READY;

    /**
     * File Status Running
     */
    static const char * FILE_STATUS_RUNNING;

    /**
     * File Status Failed
     */
    static const char * FILE_STATUS_FAILED;
    
    /**
     * File Status Done
     */
    static const char * FILE_STATUS_DONE;

    
    /**
     * Request Status Pending
     */
    static const char * REQ_STATUS_PENDING;
    
    /**
     * Request Status Ready
     */
    static const char * REQ_STATUS_READY;
    
    /**
     * Request Status Active
     */
    static const char * REQ_STATUS_ACTIVE;
    
    /**
     * Request Status Failed
     */
    static const char * REQ_STATUS_FAILED;
    
    /**
     * Request Status Failed
     */
    static const char * REQ_STATUS_DONE;
    
    /**
     * operation Context
     */
    class Context {

    public:

	typedef enum {
	    UNDEF_OPERATION = 0,
	    GET_OPERATION   = 1,
	    PUT_OPERATION   = 2,
	    COPY_OPERATION  = 3
	} OperationType;

	Context() : m_type(UNDEF_OPERATION),m_requestId(0),m_size(0) {}
    
	Context(OperationType t) : m_type(t),m_requestId(0),m_size(0) {}

	~Context() {}

	/**
	 * Get the Operation Type Property
	 * @return Operation Type Property
	 */
	OperationType getOperationType() const {return m_type;}

	/**
	 * Get the status string
	 * @return the status
	 */
	const char * getStatus()const { return m_status.c_str(); }

	/**
	 * Get the SURL Property
	 * @return the SURL property
	 */
	const char * getSURL()const { return m_surl.c_str(); }

	/**
	 * Get the TURL Property
	 * @return the TURL property
	 */
	const char * getTURL()const { return m_turl.c_str(); }

	/**
	 * Set the SURL Property
	 * @param surl [IN] the SURL value
	 */
	void setSURL(const char * surl) { m_surl = surl;}

	/**
	 * Set the File Size Property
	 * @param s [IN] the File Size value
	 */
	void setSize(LONG64 s) { m_size = s;}

	/**
	 * Get the File Size Property
	 * @return the File Size value
	 */
	LONG64 getSize() const { return m_size;}
          
	/**
	 * Set the Endpoint Property
	 * @param Endpoint [IN] the Endpoint value
	 */
	void setEndpoint(const char * e) { m_endpoint = e;}

          
	/**
	 * Set the Operation Type Property
	 * @param t [IN] the Operation Type Property
	 */
	void setOperationType(OperationType t){m_type = t;}

	/**
	 * Return the context as a string
	 */
	const char * toString();

	/**
	 * Build the context from a string, in the same format as
	 * previously emitted by toString(). Returns 0 on success, -1 on error.
	 */
	int fromString(const char *);


	/**
	 *  Get Request ID Property 
	 */
	int getRequestId() const {return m_requestId;}

	/**
	 *  Get File ID Property 
	 */
	int getFileId() const {return m_fileId;}

	/**
	 *  Get Endpoint Property 
	 */
	const char * getEndpoint() const {return m_endpoint.c_str();}


    protected:    
	friend class SrmUtil;
        
	/**
	 *  Set Request ID Property 
	 */
	void setRequestId(int reqId){m_requestId = reqId;}

	/**
	 *  Set status
	 */
	void setStatus(const char *status){m_status = status;}

	/**
	 *  Set File ID Property 
	 */
	void setFileId(int fId){m_fileId = fId;}

	/**
	 * Set the TURL Property
	 * @param surl [IN] the TURL value
	 */
	void setTURL(const char * turl) { m_turl = turl;}

	/**
	 * Constructor
	 */
	Context(const char * surl): m_surl(surl),m_requestId(0),m_fileId(0),m_type(UNDEF_OPERATION){}

    private:
	int             m_requestId;    
	int             m_fileId;    
	OperationType   m_type;
	std::string     m_surl;
	std::string     m_endpoint;
	std::string     m_turl;
	std::string     m_status;
	std::string     m_msg;
        LONG64          m_size;
    };
       
public:

    /*
     * Constructor
     */
    SrmUtil(const char *endpoint, int http_timeout = 40, const std::string * id = NULL);

    /*
     * Destructor
     */
    virtual ~SrmUtil();
    
    // -------------------------------------------------------------------------
    // SRM Common Functions
    // -------------------------------------------------------------------------
    
    /**
     * Get TURL from SURL without blocking. If it can be obtaiined immediately
     * then 0 is returned with the TURL set in the context. Otherwise 1 is returned,
     * indicating that updateTurlFromSurlRequest needs to be called to finish the process.
     * On error -1 is returned.
     */
    int requestTurlFromSurl(SrmUtil::Context * ctx, const char *protocol);

    /**
     * Get TURL from SURL following a requestTurlFromSurl call. The context
     * needs to be passed back in. It returns 1 if the request still has not finished
     * with the appropriate status string set in context. It returns 0 if the TURL may
     * be retrieved, and -1 on error.
     */
    int updateTurlFromSurlRequest(SrmUtil::Context * ctx);

    /**
     * Perform an SRM Copy call.
     */
    int doCopy(const std::vector<std::pair<std::string,std::string> > &l, SrmStatus *stat);

    /**
     * Delete a SURL
     */
    int deleteSurl(const char *surl);

    /**
     * set File Transfer Status
     */
    int setFileStatus(int reqID, int fileID, const char * status);

    /**
     * Check That SRM Service is running, invoking srmPing
     */
    int checkService(bool delegation = false);

    /**
     * Perform SRM get call
     */
    SrmStatus get(const char *surl, const char *protocol);
    
    /**
     * Perform SRM put call
     */
    SrmStatus put(const char *surl, const char *protocol, LONG64 size = 0);

    /**
     * Copy operation
     */
    SrmStatus copy(const std::vector<std::pair<std::string,std::string> > &l);

    /**
     * Retrieve updated status
     */
    SrmStatus getRequestStatus(int reqID);

    /**
     *  Get Service
     */
    ISRM *getService() const {return m_service;}

    /**
     * Retrieve last error message
     */
    const char *getErrorMsg() {return m_errormsg.c_str();}


    /**
     *  For an existing request status, retrieve the number of file statuses
     */
    static int rsGetFileStatSize(const SrmStatus s) {return s->fileStatuses->__size;}

    /**
     *  For an existing request status, retrieve the file statuses for number i
     */
    static const char* rsGetFileStat(const SrmStatus s, int i) {
	return s->fileStatuses->__ptr[i]->state;
    }


    /**
     * Finish and clean up once the request has completed successfully.
     */
    int finishRequest(SrmUtil::Context * ctx, SrmStatus reqstatp);


    /**
     * Get the endpoint which was set in the constructor
     */
    const char * endpoint() {return m_endpoint.c_str();}

protected:

    /**
     * Initialize The Srm Service
     */
    void initSrmService(bool delegation = false);

    /**
     * Finalize The Srm Service
     */
    void finiSrmService();

    /**
     * Prepare a request both for getTurlFromSurl and requestTurlFromSurl; updateTurlFromSurlRequest
     */
    SrmStatus prepareRequest(SrmUtil::Context * ctx, const char *protocol);

    /**
     * Chack that the request status returned is valid
     */
    SrmStatus SrmUtil::checkRequestStatus(SrmStatus reqstatp);

    /**
     * Handle errors, log the error string and clean up
     */
    void handleError(const char *,int id = 0);

    /**
     * Log the request ID for the given call
     */
    void logRequestID(const char *what, int id);

private:

    /**
     * The Logger
     */
    class log4cxx::LoggerPtr m_logger;

    /**
     * The endpoint
     */
    std::string m_endpoint;

    /**
     * Error message
     */
    std::string m_errormsg;

    /**
     * The service pointer
     */
    ISRM * m_service;

    /**
     * The delegation flag
     */
    bool m_delegation;

    /**
     * The timeout for http connections
     */
    int m_http_timeout;
};

} // End namespace srm

#endif // GLITE_DATA_IO_SRM_CLIENT_SRMUTIL_HXX_
